#include "object.h"
#include "object-type.h"
#include "value-content.h"
#include <stdlib.h>

size_t Object_AllocationSizeForType(ObjectType* otype)
{
	return sizeof(Object) + sizeof(union ValueContent) * ObjectType_GetAttributeCount(otype);
}

int Object_Destroy(void* arg, Object* o)
{
	(void) o;
	(void) arg;
	return EXIT_SUCCESS;
}

size_t Object_SizeQuery(void* arg, Object* o)
{
	(void) arg;
	return Object_AllocationSizeForType(o->type);
}

typedef struct AttributeVisitorContext_s {
	Object* self;
	TracingHeapVisit trace_visit;
	void* visit_context;
} AttributeVisitorContext;

static int _Object_AttributeVisitor(AttributeVisitorContext* context, char* attribute_name, size_t attribute_name_length,  ObjectTypeAttribute* value)
{
	(void) attribute_name;
	(void) attribute_name_length;
	switch (value->type) {
		case VALUETYPE_BYTE:
		case VALUETYPE_INT64:
		case VALUETYPE_DOUBLE:
		case VALUETYPE_BOOLEAN:
			break;

		case VALUETYPE_OBJECT:
		{
			union ValueContent attribute_value = context->self->attributes[value->index];
			// visit all object attributes
			int visit_code = context->trace_visit(context->visit_context, attribute_value.object);
			if (visit_code != EXIT_SUCCESS) {
				return visit_code;
			}
			break;
		}
        }
        return EXIT_SUCCESS;
}

int Object_ReferenceTrace(void* context, Object* object, TracingHeapVisit trace_callback, void* callback_context)
{
	(void) context;
	ObjectType* type = object->type;


	AttributeVisitorContext visitor_context = {
		.self = object,
		.trace_visit = trace_callback,
		.visit_context = callback_context,
	};

	while (type != NULL) {
		// object has a reference to its type
		int type_visit = trace_callback(callback_context, type);
		if (type_visit != EXIT_SUCCESS) {
			return type_visit;
		}

		int hashmap_visit = HashMap_ForEach(&type->attributes, (HashMapVisitFunction) _Object_AttributeVisitor, &visitor_context);
		if (hashmap_visit != EXIT_SUCCESS) {
			return hashmap_visit;
		}

		type = type->supertype;
	}


	return EXIT_SUCCESS;
}
